extern "C" {
#define XMD_H
#include "jinclude.h"
#include "jpeglib.h"
}

#include "playcap.h"
#include <math.h>
#include <Vfw.h>
#include "resource.h"
#include "dd.h"
#include <commctrl.h>

#include <vector>

class mycapture : public CameraCapture {
public:
    enum filtertype {
	FILTER_NONE,
	FILTER_BLUR,
	FILTER_SWAPREDBLUE,
	FILTER_DIFFUSE,
	FILTER_INVERSE,
	FILTER_TIGHTROBERT,
	FILTER_ROBERT,
	FILTER_EMBOSS,
	FILTER_PAPER,
	FILTER_LEPROSY,
	FILTER_EDGEDETECTFROMMOTIONDETECTOR,
	FILTER_TEST
    };

    mycapture(DirectDrawWin *dd) {
	_ddwin = dd;
	_movement_edge1 = NULL;
	_movement_edge2 = NULL;
	_effect = FILTER_NONE;
	_orig_dst = NULL;
	_motion_only = 0;

	// how many pixels need to change to signify motion
	_motion_threshold = 0.0012;

	// out of 255, the value to regard as edge
	_edge_threshold = 110;
    }

    ~mycapture() {
	delete [] _movement_edge1;
	delete [] _movement_edge2;
    }


    void set_effect(enum filtertype e) { _effect = e; }
    void motion_only(bool m) { _motion_only = m; }
    bool motion_only(void) { return _motion_only; }
    void motion_threshold(double d) { _motion_threshold = d; }
    double motion_threshold(void) { return _motion_threshold; }
    void grabframe(const char *filename) { _grabframe = filename; }

private:
    DirectDrawWin *_ddwin;

    u_char *_orig_dst;
    u_char *_movement_edge1;
    u_char *_movement_edge2;

    std::string _grabframe;

    enum filtertype _effect;
    bool _motion_only;
    double _motion_threshold;
    int _edge_threshold;

    bool has_movement_occured(u_char *buf);
    void new_frame(double SampleTime, BYTE *buffer, long len);
};

class AVIFileWriter {
    PAVIFILE _avifile;
    AVISTREAMINFO _stream_hdr;

    PAVISTREAM _vid_stream;
    PAVISTREAM _vid_compressedstream;

    int _cnt;

    CRITICAL_SECTION _crit;

    class Lock {
	CRITICAL_SECTION *_c;
    public:
	Lock(CRITICAL_SECTION *c) {
	    _c = c;
	    EnterCriticalSection(_c);
	}
	~Lock() {
	    LeaveCriticalSection(_c);
	}
    };

public:
    AVIFileWriter() {
	InitializeCriticalSection(&_crit);
	_vid_stream = NULL;
	_vid_compressedstream = NULL;
	_avifile = NULL;
	_cnt = 0;
    }
    ~AVIFileWriter() {
	close();
	DeleteCriticalSection(&_crit);
    }

    void close(void) {
	Lock _lock(&_crit);
	if (_vid_stream) AVIStreamClose(_vid_stream);
	if (_vid_compressedstream) AVIStreamClose(_vid_compressedstream);
	if (_avifile) AVIFileClose(_avifile);
	_vid_stream = NULL;
	_vid_compressedstream = NULL;
	_avifile = NULL;
    }

    bool create(const char *filename, int rate, int w, int h, BITMAPINFO *bmi) {
	Lock _lock(&_crit);
	_vid_stream = NULL;
	_avifile = NULL;
	_cnt = 0;

	if (AVIFileOpen(&_avifile, filename, OF_CREATE | OF_WRITE, 0)) {
	    MessageBox(NULL, "could not create avi file\n",
		       "error", MB_OK|MB_ICONERROR);
	    return 0;
	}

	memset(&_stream_hdr, 0, sizeof(_stream_hdr));
	_stream_hdr.fccType = streamtypeVIDEO;
	_stream_hdr.fccHandler = 0;
	_stream_hdr.dwScale = 1;
	_stream_hdr.dwRate = rate;
	_stream_hdr.dwSuggestedBufferSize = w * h * 3;
	SetRect(&_stream_hdr.rcFrame, 0, 0, w, h);

	// And create the stream;
	if (AVIFileCreateStream(_avifile, &_vid_stream, &_stream_hdr)) {
	    MessageBox(NULL, "could not create avi stream\n",
		       "error", MB_OK|MB_ICONERROR);
	    return 0;
	}

	AVICOMPRESSOPTIONS opts;
	AVICOMPRESSOPTIONS *aopts[1] = {&opts};
	memset(&opts, 0, sizeof(opts));

	if (!AVISaveOptions(NULL, 0, 1, &_vid_stream,
			    (LPAVICOMPRESSOPTIONS *)&aopts)) {
	    MessageBox(NULL, "could not avi save options\n",
		       "error", MB_OK|MB_ICONERROR);
	    return 0;
	}

	if (AVIMakeCompressedStream(&_vid_compressedstream, _vid_stream, &opts, NULL)) {
	    MessageBox(NULL, "could not make avi compressed stream\n",
		       "error", MB_OK|MB_ICONERROR);
	    return FALSE;
	}

	HRESULT hr;
	if ((hr=AVIStreamSetFormat(_vid_compressedstream, 0, bmi,
				   bmi->bmiHeader.biSize +
				   bmi->bmiHeader.biClrUsed * sizeof(RGBQUAD)))) {
	    std::string s = "could not set format of avi compressed stream: ";

	    if (hr == AVIERR_NOCOMPRESSOR)
		s += "AVIERR_NOCOMPRESSOR";
	    else if (hr == AVIERR_MEMORY)
		s += "AVIERR_MEMORY";
	    else if (hr == AVIERR_UNSUPPORTED)
		s += "AVIERR_UNSUPPORTED";
	    else {
		char buf[100];
		sprintf(buf, "0x%x", hr);
		s += "OTHER_ERROR(";
		s += buf;
		s += ")";
	    }

	    s += "\n";
	    MessageBox(NULL, s.c_str(), "error", MB_OK|MB_ICONERROR);
	    return FALSE;
	}

	AVISaveOptionsFree(1, (LPAVICOMPRESSOPTIONS *)&aopts);

	return 1;
    }

    bool writevidframe(void *data, LONG sz, BOOL iskey) {
	Lock _lock(&_crit);
	if (_vid_compressedstream == NULL)
	    return false;

	HRESULT hr;
	if ((hr=AVIStreamWrite(_vid_compressedstream,
			       _cnt, 1, data, sz,
			       iskey ? AVIIF_KEYFRAME : 0, NULL, NULL))) {
	    printf("ERROR 0x%x\n", hr);
	    fflush(stdout);
	    MessageBox(NULL, "could not write video frame to avi file\n",
		       "error", MB_OK|MB_ICONERROR);
	    return false;
	}
	_cnt++;
	return true;
    }
};

AVIFileWriter aviwriter;
DirectDrawWin ddwin;
mycapture *vid;

LRESULT CALLBACK
vidWndProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
    switch (uMsg) {
	case WM_PAINT:
	    {
		PAINTSTRUCT ps;
		BeginPaint(hwnd, &ps);
		ddwin.draw();
		EndPaint(hwnd, &ps);
	    }
	    return 0;
    }

    return DefWindowProc(hwnd, uMsg, wParam, lParam);
}

void uncheckallfilters(HWND hwnd)
{
#define XX(x) CheckMenuItem(GetMenu(hwnd), ID_FILTER_ ## x, \
			    MF_BYCOMMAND | MF_UNCHECKED)
    XX(NONE);
    XX(BLUR);
    XX(SWAPREDBLUE);
    XX(DIFFUSE);
    XX(INVERSE);
    XX(TIGHTROBERT);
    XX(ROBERT);
    XX(EMBOSS);
    XX(PAPER);
    XX(LEPROSY);
    XX(EDGEDETECTFROMMOTIONDETECTOR);
    XX(TEST);
#undef XX
}

INT_PTR CALLBACK
aboutproc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    switch (message) {
	case WM_COMMAND:
	    switch (LOWORD(wParam)) {
		case IDOK:
		case IDCANCEL:
		    EndDialog(hwnd, LOWORD(wParam));
	    }
	    return TRUE;
    }
    return FALSE;
}
std::vector<std::string> *vidsellist;

INT_PTR CALLBACK
devselproc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    static int *ret;
    switch (message) {
	case WM_INITDIALOG:
	    {
		ret = (int*)lParam;
		HWND hList = GetDlgItem(hwnd, IDC_LIST);

		int i = LVS_EX_FULLROWSELECT;
#ifdef LVS_EX_LABELTIP
		i |= LVS_EX_LABELTIP;
#endif
		ListView_SetExtendedListViewStyle(hList, i);

		LVCOLUMN lvc;
		lvc.mask = LVCF_FMT | LVCF_WIDTH | LVCF_TEXT ;
		lvc.cx = 223;
		lvc.fmt = LVCFMT_LEFT;
		lvc.pszText = "Video Sources";
		ListView_InsertColumn(hList, 0, &lvc);

		i = 0;
		std::vector<std::string>::iterator it = vidsellist->begin();
		while (it != vidsellist->end()) {
		    LVITEM lvi;
		    ZeroMemory(&lvi, sizeof(lvi));
		    lvi.mask = LVIF_IMAGE | LVIF_PARAM | LVIF_TEXT;
		    lvi.iItem = i;
		    lvi.pszText = (char *)it->c_str();
		    lvi.lParam = i;
		    lvi.iSubItem = 0;
		    lvi.iImage = i;

		    ListView_InsertItem(hList, &lvi);

		    i++;
		    it++;
		}
		ListView_SetItemState(hList, 0, LVIS_SELECTED, LVIS_SELECTED);
	    }
	    SetFocus(GetDlgItem(hwnd, IDOK));
	    return FALSE; // this is because we set focus

	case WM_COMMAND:
	    switch (LOWORD(wParam)) {
		case IDOK:
		    {
			HWND hList = GetDlgItem(hwnd, IDC_LIST);
			*ret = ListView_GetNextItem(hList, -1, LVNI_SELECTED);
			EndDialog(hwnd, IDOK);
		    }
		    break;
		case IDCANCEL:
		    EndDialog(hwnd, LOWORD(wParam));
		    break;
	    }
	    return TRUE;
    }
    return FALSE;
}

int
SelectDevice(std::vector<std::string> &v)
{
    vidsellist = &v;
    int i;
    switch (DialogBoxParam(GetModuleHandle(NULL),
			   MAKEINTRESOURCE(IDD_DEVICESELECT),
			   NULL, reinterpret_cast<DLGPROC>(devselproc),
			   (LPARAM)&i)) {
	case IDOK:
	    return i;
	default:
	    return -1;
    }
}

int CALLBACK
dlgfunc(HWND hwndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
    switch (uMsg) {
	case WM_CLOSE:
	    delete vid;
	    aviwriter.close();
	    break;

	case WM_MOVE:
	    ddwin.draw();
	    return 0;

	case WM_INITDIALOG:
	    {
		vid = new mycapture(&ddwin);
		std::vector<std::string> v;
		if (!vid->FindCaptureDevices(v)) {
		    MessageBox(NULL, vid->GetError(),
			       "Error: Could not get devices", MB_OK|MB_ICONERROR);
		    PostQuitMessage(0);
		    return 0;
		}
		int choice = 0;
		if (v.size() > 1)
		    choice = SelectDevice(v);
		if (choice < 0) {
		    delete vid;
		    vid = NULL;
		    PostQuitMessage(0);
		    return 0;
		}
		if (!vid->Init(v[choice].c_str())) {
		    MessageBox(NULL, vid->GetError(),
			       "Error: Could not init video", MB_OK|MB_ICONERROR);
		    PostQuitMessage(0);
		    return 0;
		}
	    }

	    HWND hwndVid;
	    {
		static char *szAppName = "vidfilt_video";
		WNDCLASS     wndclass;
		wndclass.style         = 0;
		wndclass.lpfnWndProc   = vidWndProc;
		wndclass.cbClsExtra    = 0;
		wndclass.cbWndExtra    = 0;
		wndclass.hInstance     = GetModuleHandle(NULL);
		wndclass.hIcon         = LoadIcon (NULL, IDI_APPLICATION);
		wndclass.hCursor       = LoadCursor (NULL, IDC_ARROW);
		wndclass.hbrBackground = (HBRUSH) GetStockObject (WHITE_BRUSH);
		wndclass.lpszMenuName  = NULL;
		wndclass.lpszClassName = szAppName;
		RegisterClass(&wndclass);

		hwndVid = CreateWindow(szAppName, "", WS_CHILD|WS_VISIBLE,
				       0, 0, vid->Width(), vid->Height(),
				       hwndDlg,
				       NULL, GetModuleHandle(NULL), NULL);

		RECT r;
		GetClientRect(hwndDlg, &r);
		if (r.right - 40 < vid->Width())
		    r.right = vid->Width() + 40;
		if (r.bottom - 40 < vid->Height())
		    r.bottom = vid->Height() + 40;

		SetWindowPos(hwndDlg, HWND_TOP, 0, 0, r.right, r.bottom + 40, SWP_NOMOVE);
		SetWindowPos(hwndVid, HWND_TOP,
			     (r.right - vid->Width()) / 2 ,
			     (r.bottom - vid->Height()) / 2 ,
			     vid->Width(), vid->Height(), 0);
		UpdateWindow(hwndVid);
	    }

	    if (!ddwin.init(hwndVid)) {
		MessageBox(NULL, "could not init directdraw\n",
			   "error", MB_OK|MB_ICONERROR);
		PostQuitMessage(0);
	    }

	    vid->Start();

	    return false;

	case WM_COMMAND:
	    switch (LOWORD(wParam)) {
		case IDCANCEL:
		    EndDialog(hwndDlg, wParam);
		    return true;
#define XX(x) case ID_FILTER_ ## x : uncheckallfilters(hwndDlg); \
		    vid->set_effect(mycapture::FILTER_ ## x); \
		    CheckMenuItem(GetMenu(hwndDlg), ID_FILTER_ ## x, \
					  MF_BYCOMMAND | MF_CHECKED); break


		    XX(NONE);
		    XX(BLUR);
		    XX(SWAPREDBLUE);
		    XX(DIFFUSE);
		    XX(INVERSE);
		    XX(TIGHTROBERT);
		    XX(ROBERT);
		    XX(EMBOSS);
		    XX(PAPER);
		    XX(LEPROSY);
		    XX(EDGEDETECTFROMMOTIONDETECTOR);
		    XX(TEST);
#undef XX

		case ID_SETTINGS_ONLYCAPTUREMOTION:
		    {

			if (GetMenuState(GetMenu(hwndDlg),
			    ID_SETTINGS_ONLYCAPTUREMOTION, MF_BYCOMMAND) &
			    MF_CHECKED) {
			    vid->motion_only(0);
			    CheckMenuItem(GetMenu(hwndDlg),
					  ID_SETTINGS_ONLYCAPTUREMOTION,
					  MF_BYCOMMAND | MF_UNCHECKED);
			} else {
			    vid->motion_only(1);
			    CheckMenuItem(GetMenu(hwndDlg),
					  ID_SETTINGS_ONLYCAPTUREMOTION,
					  MF_BYCOMMAND | MF_CHECKED);
			}
		    }
		    return false;

		case ID_HELP_ABOUT:
		    DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_ABOUT),
			      hwndDlg, aboutproc);
		    break;

		case ID_CAPTURE_PREVIEW:
		    CheckMenuItem(GetMenu(hwndDlg), ID_CAPTURE_PREVIEW,
					  MF_BYCOMMAND | MF_CHECKED);
		    CheckMenuItem(GetMenu(hwndDlg), ID_CAPTURE_CAPTURETOFILE,
					  MF_BYCOMMAND | MF_UNCHECKED);
		    aviwriter.close();
		    break;

		case ID_CAPTURE_CAPTURETOFILE:
		    if (GetMenuState(GetMenu(hwndDlg),
				     ID_CAPTURE_CAPTURETOFILE, MF_BYCOMMAND) &
			MF_CHECKED)
		    {
			CheckMenuItem(GetMenu(hwndDlg), ID_CAPTURE_PREVIEW,
				      MF_BYCOMMAND | MF_CHECKED);
			CheckMenuItem(GetMenu(hwndDlg), ID_CAPTURE_CAPTURETOFILE,
				      MF_BYCOMMAND | MF_UNCHECKED);
			aviwriter.close();

		    } else {
			CheckMenuItem(GetMenu(hwndDlg), ID_CAPTURE_PREVIEW,
				      MF_BYCOMMAND | MF_UNCHECKED);
			CheckMenuItem(GetMenu(hwndDlg), ID_CAPTURE_CAPTURETOFILE,
				      MF_BYCOMMAND | MF_CHECKED);
			{
			    OPENFILENAME sv;
			    char defaultfile[1024];
			    strcpy(defaultfile, "movie.avi");
			    memset(&sv, 0, sizeof(sv));
			    sv.lStructSize       = sizeof(sv);
			    sv.hwndOwner         = hwndDlg;
			    sv.hInstance         = GetModuleHandle(NULL);
			    sv.lpstrFile         = defaultfile;
			    sv.nMaxFile          = sizeof(defaultfile);
			    sv.Flags             = OFN_HIDEREADONLY |
				OFN_OVERWRITEPROMPT;
			    sv.lpstrDefExt       = "avi";
			    sv.lpstrFilter       = "AVI files (*.avi)\0*.AVI\0\0";

			    if (GetSaveFileName(&sv)) {
				BITMAPINFO bmi;
				bmi.bmiHeader.biSize = sizeof(bmi.bmiHeader);
				bmi.bmiHeader.biWidth = vid->Width();
				bmi.bmiHeader.biHeight = vid->Height();
				bmi.bmiHeader.biPlanes = 1;
				bmi.bmiHeader.biBitCount = 24;
				bmi.bmiHeader.biCompression = BI_RGB;
				bmi.bmiHeader.biSizeImage =
				    bmi.bmiHeader.biWidth * bmi.bmiHeader.biHeight * 3;
				bmi.bmiHeader.biXPelsPerMeter = 0;
				bmi.bmiHeader.biYPelsPerMeter = 0;
				bmi.bmiHeader.biClrUsed = 0;
				bmi.bmiHeader.biClrImportant = 0;
				aviwriter.create(defaultfile,
						 15, // fps
						 vid->Width(), vid->Height(),
						 &bmi);
			    }
			}
		    }
		    return true;

		case ID_CAPTURE_CAPTUREFRAME:
		    {
			OPENFILENAME sv;
			char defaultfile[1000];
			strcpy(defaultfile, "snapshot.jpg");
			memset(&sv, 0, sizeof(sv));
			sv.lStructSize       = sizeof(sv);
			sv.hwndOwner         = hwndDlg;
			sv.hInstance         = GetModuleHandle(NULL);
			sv.lpstrFile         = defaultfile;
			sv.nMaxFile          = sizeof(defaultfile);
			sv.Flags             = OFN_HIDEREADONLY |
			    OFN_OVERWRITEPROMPT;
			sv.lpstrDefExt       = "jpg";
			sv.lpstrFilter       = "JPEG files (*.jpg)\0*.JPG\0\0";

			if (GetSaveFileName(&sv)) {
			    vid->grabframe(defaultfile);
			}
		    }
		    return true;
	    }
    }
    return false;
}


int APIENTRY WinMain(HINSTANCE hInstance,
		     HINSTANCE hPrevInstance,
		     LPSTR     lpCmdLine,
		     int       nCmdShow)
{
    WORD wVer = HIWORD(VideoForWindowsVersion());
    if (wVer < 0x010a){
	MessageBeep(MB_ICONHAND);
	MessageBox(NULL, "Sorry, Can't init AVI lib !", "Error",
		   MB_OK|MB_ICONSTOP);
	return 0;
    }

    AVIFileInit();

    DialogBox(hInstance, MAKEINTRESOURCE(IDD_MAINDLG), NULL, dlgfunc);

    return 0;
}


int
get_pixel(u_char *p, u_char *r, u_char *g, u_char *b)
{
    *r = p[2]; *g = p[1]; *b = p[0];
    return 3;
}

void
set_pixel(u_char *p, u_char r, u_char g, u_char b)
{
    p[2] = r;
    p[1] = g;
    p[0] = b;
}

void
mycapture::new_frame(double SampleTime, BYTE *buffer, long len)
{
    int w = Width();
    int h = Height();

    int pos[9] = {
	(-3*w-3), (-3*w), (-3*w+3),
	(-3), (0), (3),
	(+3*w-3), (+3*w-0), (+3*w+3)
    };

    u_char *src, *orig_src, *dst;
    orig_src = src = buffer;

    if (_orig_dst == NULL) {
	_orig_dst = new u_char[w * h * 3];
	_movement_edge1 = new u_char[w * h];
	_movement_edge2 = new u_char[w * h];
    }
    dst = _orig_dst;

    int pitch;
    u_char *ddbuf = _ddwin->GetBuffer(&pitch);
    ddbuf += pitch * (h-1);

#define SETPIX(r, g, b) do {\
    ddb += _ddwin->set_pixel(ddb, (u_char)r, (u_char)g, (u_char)b); \
    set_pixel(dst, (u_char)r, (u_char)g, (u_char)b); \
} while (0)

    int y = 0;
    while (y < h)
    {
	u_char *ddb = ddbuf;
	int x = 0;
	while (x < w) {
	    if (x == 0 || y == 0 || x == w-1 || y == h-1)
	    {
		SETPIX(0, 0, 0);
	    }
	    else
	    {
		if (_effect == FILTER_NONE)
		{
		    // nothing

		    u_char r, g, b;
		    get_pixel(src, &r, &g, &b);
		    SETPIX(r, g, b);
		}
		else if (_effect == FILTER_BLUR)
		{
		    // blur

		    int r1 = 0, g1 = 0, b1 = 0;
		    int i = 0;
		    u_char r, g, b;

		    for (; i < 9; i++) {
			get_pixel(src+pos[i], &r, &g, &b);
			r1 += r;
			g1 += g;
			b1 += b;
		    }
		    r1 /= 9;
		    g1 /= 9;
		    b1 /= 9;
		    SETPIX(r1, g1, b1);
		}
		else if (_effect == FILTER_SWAPREDBLUE)
		{
		    // swap red and blue

		    u_char r, g, b;
		    get_pixel(src, &r, &g, &b);
		    SETPIX(b, g, r);
		}
		else if (_effect == FILTER_DIFFUSE)
		{
		    // diffuse

		    u_char *s = src+pos[rand() % 9];
		    u_char r, g, b;
		    get_pixel(s, &r, &g, &b);
		    SETPIX(r, g, b);
		}
		else if (_effect == FILTER_INVERSE)
		{
		    // inverse

		    u_char r, g, b;
		    get_pixel(src, &r, &g, &b);
		    SETPIX(~r, ~g, ~b);
		}
		else if (_effect == FILTER_TIGHTROBERT)
		{
		    // tight roberts cross

		    int delta_r1, delta_g1, delta_b1;
		    int delta_r2, delta_g2, delta_b2;

		    u_char r1, g1, b1;
		    u_char r2, g2, b2;

		    get_pixel(src, &r1, &g1, &b1);
		    get_pixel(src+pos[8], &r2, &g2, &b2);
		    delta_r1 = r1 - r2;
		    delta_g1 = g1 - g2;
		    delta_b1 = b1 - b2;

		    get_pixel(src+pos[2], &r2, &g2, &b2);
		    delta_r2 = r1 - r2;
		    delta_g2 = g1 - g2;
		    delta_b2 = b1 - b2;

		    if (delta_r1 < 0) delta_r1 = -delta_r1;
		    if (delta_g1 < 0) delta_g1 = -delta_g1;
		    if (delta_b1 < 0) delta_b1 = -delta_b1;
		    if (delta_r2 < 0) delta_r2 = -delta_r2;
		    if (delta_g2 < 0) delta_g2 = -delta_g2;
		    if (delta_b2 < 0) delta_b2 = -delta_b2;

		    SETPIX(delta_r1 + delta_r2,
			   delta_g1 + delta_g2,
			   delta_b1 + delta_b2);
		}
		else if (_effect == FILTER_ROBERT || _effect == FILTER_PAPER)
		{
		    // roberts cross

		    int delta_r1, delta_g1, delta_b1;
		    int delta_r2, delta_g2, delta_b2;

		    u_char r1, g1, b1;
		    u_char r2, g2, b2;

		    get_pixel(src+pos[0], &r1, &g1, &b1);
		    get_pixel(src+pos[8], &r2, &g2, &b2);
		    delta_r1 = r1 - r2;
		    delta_g1 = g1 - g2;
		    delta_b1 = b1 - b2;

		    get_pixel(src+pos[6], &r1, &g1, &b1);
		    get_pixel(src+pos[2], &r2, &g2, &b2);
		    delta_r2 = r1 - r2;
		    delta_g2 = g1 - g2;
		    delta_b2 = b1 - b2;

		    if (delta_r1 < 0) delta_r1 = -delta_r1;
		    if (delta_g1 < 0) delta_g1 = -delta_g1;
		    if (delta_b1 < 0) delta_b1 = -delta_b1;
		    if (delta_r2 < 0) delta_r2 = -delta_r2;
		    if (delta_g2 < 0) delta_g2 = -delta_g2;
		    if (delta_b2 < 0) delta_b2 = -delta_b2;

		    if (_effect == FILTER_PAPER) {
			int i = (delta_r1 + delta_r2,
				 delta_g1 + delta_g2,
				 delta_b1 + delta_b2) / 3;

			if (i < 0x40)
			    i *= 4;

			if (i > 0x80)
			    i = 0;
			else
			    i = ~i;

			SETPIX(i, i, i);
		    } else {
			SETPIX(delta_r1 + delta_r2,
			       delta_g1 + delta_g2,
			       delta_b1 + delta_b2);
		    }
		}
		else if (_effect == FILTER_EMBOSS)
		{
		    // emboss

		    int r, g, b;
		    u_char r0, g0, b0;
		    u_char r1, g1, b1;
		    u_char r3, g3, b3;
		    u_char r4, g4, b4;
		    u_char r5, g5, b5;
		    u_char r7, g7, b7;
		    u_char r8, g8, b8;

		    get_pixel(src+pos[0], &r0, &g0, &b0);
		    get_pixel(src+pos[1], &r1, &g1, &b1);
		    get_pixel(src+pos[3], &r3, &g3, &b3);
		    get_pixel(src       , &r4, &g4, &b4);
		    get_pixel(src+pos[5], &r5, &g5, &b5);
		    get_pixel(src+pos[7], &r7, &g7, &b7);
		    get_pixel(src+pos[8], &r8, &g8, &b8);

		    r = 0x80 + r0 - r4 + r4 - r8 + r5 - r3 + r1 - r7;
		    r = (r < 0) ? 0 : (r > 0xff ? 0xff : r);

		    g = 0x80 + g0 - g4 + g4 - g8 + g5 - g3 + g1 - g7;
		    g = (g < 0) ? 0 : (g > 0xff ? 0xff : g);

		    b = 0x80 + b0 - b4 + b4 - b8 + b5 - b3 + b1 - b7;
		    b = (b < 0) ? 0 : (b > 0xff ? 0xff : b);

		    SETPIX(r, g, b);
		}
		else if (_effect == FILTER_LEPROSY) // effect 8 is part of effect 6
		{
		    // leprosy
		    if (x < 5 || x > w-5 || y < 5 || y > h-5) {
		    } else {
			u_char _r1, _g1, _b1;
			u_char _r2, _g2, _b2;
			get_pixel(src, &_r1, &_g1, &_b1);
			get_pixel(src+(pos[1]*5), &_r2, &_g2, &_b2);

			u_long r1 = _r1, g1 = _g1, b1 = _b1;
			int r2 = _r2, g2 = _g2, b2 = _b2;

			if ((r2 - r1) > 32) r1 -= 32;
			if ((g2 - g1) > 32) g1 -= 32;
			if ((b2 - b1) > 32) b1 -= 32;

			if (r1 > 255) r1 = 0;
			if (g1 > 255) g1 = 0;
			if (b1 > 255) b1 = 0;

			SETPIX((u_char)r1, (u_char)g1, (u_char)b1);
		    }
		}
		else if (_effect == FILTER_TEST)
		{
		    static double rc = 0.0;
		    static double gc = 0.0;

		    u_char r, g, b;
		    get_pixel(src, &r, &g, &b);

		    long r1 = r;
		    long g1 = g;

		    r1 += (long)(sin(rc) * 80);
		    g1 += (long)(cos(gc) * 80);

		    if (r1 < 0) r1 = 0;
		    if (g1 < 0) g1 = 0;
		    if (r1 > 255) r1 = 255;
		    if (g1 > 255) g1 = 255;

		    SETPIX(r1, g1, b);

		    rc += 0.11;
		    gc += 0.85;
		}
		else if (_effect == FILTER_EDGEDETECTFROMMOTIONDETECTOR)
		{
		    double sum1, sum2, sum;

		    u_char r0, g0, b0;
		    u_char r1, g1, b1;
		    u_char r2, g2, b2;
		    u_char r3, g3, b3;
		    u_char r4, g4, b4;
		    u_char r5, g5, b5;
		    u_char r6, g6, b6;
		    u_char r7, g7, b7;
		    u_char r8, g8, b8;

		    get_pixel(src+pos[0], &r0, &g0, &b0);
		    get_pixel(src+pos[1], &r1, &g1, &b1);
		    get_pixel(src+pos[2], &r2, &g2, &b2);
		    get_pixel(src+pos[3], &r3, &g3, &b3);
		    get_pixel(src       , &r4, &g4, &b4);
		    get_pixel(src+pos[5], &r5, &g5, &b5);
		    get_pixel(src+pos[6], &r6, &g6, &b6);
		    get_pixel(src+pos[7], &r7, &g7, &b7);
		    get_pixel(src+pos[8], &r8, &g8, &b8);

		    sum1 = (double)(r2) - (double)(r0) +
			2.0 * ( (double)(r5) - (double)(r3) ) +
			(double)(r8) - (double)(r6);

		    sum2 = ( (double)(r6) + 2.0 * (double)(r7) +
			     (double)(r8) ) -
			( (double)(r0) + 2.0 * (double)(r1) +
			  (double)(r2) );
		    sum = sqrt( sum1 * sum1 + sum2 * sum2 );
		    sum /= 1.4; // arbitrary scaling factor
		    if (sum > 255.0) sum = 255.0;

		    SETPIX(sum, sum, sum);
		}
	    }
	    src += 3;
	    dst += 3;
	    x++;
	}
	y++;
	ddbuf -= pitch;
    }

    _ddwin->ReleaseBuffer();

    // this is because has_movement_occured() might move around the buffers
    if (_motion_only == 0 || has_movement_occured(_orig_dst)) {
	aviwriter.writevidframe(_orig_dst, len, 1);
	if (!_grabframe.empty()) {
	    FILE *out = fopen(_grabframe.c_str(), "wb");
	    _grabframe = "";

	    // save frame to jpeg
	    struct jpeg_compress_struct cinfo;
	    struct jpeg_error_mgr jerr;

	    cinfo.err = jpeg_std_error(&jerr);
	    jpeg_create_compress(&cinfo);
	    jpeg_stdio_dest(&cinfo, out);
	    cinfo.image_width = Width();
	    cinfo.image_height = Height();
	    cinfo.input_components = 3;
	    cinfo.in_color_space = JCS_RGB;
	    cinfo.optimize_coding = 1;
	    jpeg_set_defaults(&cinfo);
	    jpeg_set_quality (&cinfo, 75, TRUE );
	    jpeg_start_compress(&cinfo, TRUE);

	    u_char *buf = new u_char[Width() * 3];
	    int i = 0;
	    unsigned char *line = _orig_dst + ((Width() * 3) * (Height()-1));
	    while (i < Height()) {
		u_char *s = line;
		u_char *d = buf;
		int j = 0;
		while (j < Width()) {
		    d[0] = s[2];
		    d[1] = s[1];
		    d[2] = s[0];

		    d += 3;
		    s += 3;
		    j++;
		}
		jpeg_write_scanlines(&cinfo, &buf, 1);
		i++;
		line -= Width()*3;
	    }
	    jpeg_finish_compress(&(cinfo));
	    jpeg_destroy_compress(&(cinfo));
	    delete [] buf;

	    fclose(out);
	}
	_ddwin->draw();
    }
}

void
mkedge(u_char *pic, u_char *epic, int w, int h)
{
/*
    x = @3 - @1                  1 2 3
      + @9 - @7                  4 5 6
      + 2*(@6 - @4)              7 8 9

    y = @7 + 2*(@8) + @9
      - @1 + 2*(@2) + @3

    z = sqrt(x^2 + y^2)
    z /= scalefactor (arbitrary scaling factor)
*/

    int pos[9] = {
	(-3*w-3), (-3*w), (-3*w+3),
	(-3), (0), (3),
	(+3*w-3), (+3*w-0), (+3*w+3)
    };

    int row, col;

    u_char *out = epic;
    u_char *in = pic + (w * 3);

    // no edges on top
    for (col = 0; col < w; col++) *out++ = 0;

    // lets do body
    for (row = 1; row < h - 1; row++) {
	*out++ = 0; // first column has no edge

	in += 3;

	for (col = 1; col < w - 2; col++) {
	    double sum, sum1, sum2;

	    sum1 = (double)*(in+pos[2]) - (double)*(in+pos[0]) +
		    2.0 * ( (double)*(in+pos[5]) - (double)*(in+pos[3]) ) +
		    (double)*(in+pos[8]) - (double)*(in+pos[6]);

	    sum2 = ( (double)*(in+pos[6]) + 2.0 * (double)*(in+pos[7]) +
		     (double)*(in+pos[8]) ) -
		( (double)*(in+pos[0]) + 2.0 * (double)*(in+pos[1]) +
		  (double)*(in+pos[2]) );
	    sum = sqrt( sum1 * sum1 + sum2 * sum2 );
	    sum /= 1.4; // arbitrary scaling factor
	    if (sum > 255.0) sum = 255.0;

	    *out++ = (u_char)sum;
	    in += 3;
	}
	in += 3;
	*out++ = 0; // last column has no edge
    }

    // no edges on bottom
    for (col = 0; col < w; col++) *out++ = 0;
}

bool
mycapture::has_movement_occured(u_char *buf)
{
    bool ret = 0;
    mkedge(buf, _movement_edge1, Width(), Height());

    size_t changed_pixels = 0;
    u_char *src = _movement_edge1, *dst = _movement_edge2;

    size_t sz = Width() * Height();
    size_t thres = (size_t)(sz * _motion_threshold);

    for (size_t i = 0; i < sz; i++) {
	int c = *src - *dst;
	if (c < 0) c = -c;
	if (c > _edge_threshold) changed_pixels++;
	if (changed_pixels > thres) {
	    ret = 1;
	    break;
	}
	dst++;
	src++;
    }

    // swap the buffers for next time
    u_char *tmp = _movement_edge1;
    _movement_edge1 = _movement_edge2;
    _movement_edge2 = tmp;

    return ret;
}
